# Copyright(C) 2019 VMware Inc.

import contextlib
from copy import deepcopy
import json
import logging

import time
from urllib import parse
from typing import Sequence  # noqa

import requests

from scripts import csp_lib

"""Cerebro Spring library support.

Minimal usage example::
    from vmware_symphony_deploy import cerebro_lib
    with requests.Session() as session:
        csp_uri= "https://api.testing.symphony-dev.com"
        def get_token():
            return cerebro_spring_lib.get_csp_auth_token(
                session,
                csp_uri,
                cerebro_spring_lib.AUTOMATION_USER,
                cerebro_spring_lib.AUTOMATION_CSP_SECRET)

    adapter = cerebro_spring_lib.KubernetesProxyAdapter(
        pkapi, 'testing', "cerebro-service", "cerebro-http", get_token)

    cerebro_session = cerebro_spring_lib.cerebro_session(adapter)
    cerebro = cerebro_spring_lib.Cerebro(cerebro_session)
    for shard in cerebro.shards():
        print(shard)

    shard_obj = {
        "serviceName": "robert-test",
        "address": "arbitrarytext",
        "labels": {"weight": "0"}}
    shard = cerebro_spring_lib.Shard(cerebro_session, shard_obj)
    print(shard.create())
"""

AUTOMATION_USER = "cerebro-automation"
AUTOMATION_CSP_SECRET = "OU3B#B@MO4*^$Kst"


class KubernetesProxyAdapter(requests.adapters.HTTPAdapter):
    """A HTTP API interface for making HTTP requests via kubernetes."""

    def __init__(self, api, namespace, service, port, token_cb):
        """Create a KubernetesHTTPHandle.

        The hostname and port in requests is ignored, and all requests are
        directed to service and port.

        :param api: A pykube API handle.
        :param namespace: The k8s namespace cerebro is running within.
        :param service: The service to connect to.
        :param port: The port to conenct to.
        :param token_cb: A callback to get a token for making Cerebro operator
            calls.
        """
        self.api = api
        self.namespace = namespace
        self.token_cb = token_cb
        self.service = service
        self.port = port
        self.token = None
        super().__init__()

    def send(self, request, **kwargs):
        if self.token is None and self.token_cb:
            self.token = self.token_cb()

        if self.token:
            request.headers["x-xenon-auth-token"] = self.token
        request.headers.pop("Authorization", None)
        # http://kubernetes_master_address/api/v1/namespaces/namespace_name/services/service_name[:port_name]/proxy
        k8surl = "services/%(service)s:%(port)s/proxy/" % dict(
            service=self.service, port=self.port
        )
        url_s = parse.urlsplit(request.url)
        url = parse.urlunsplit(
            ("http", "", parse.urljoin(k8surl, url_s[2][1:]), url_s[3], url_s[4])
        )[8:]

        kwargs = dict(url=url, namespace=self.namespace, headers=request.headers)
        if request.method in ("POST", "PUT", "PATCH"):
            kwargs["data"] = request.body
        return getattr(self.api, request.method.lower())(**kwargs)


def cerebro_session(adapter):
    """Construct a requests session that will submit requests to cerebro.

    :param adapter: A KubernetesProxyAdapter for submitting requests.
    :return: A requests.Session() where requests to http://anyhost/ will be
        submitted via the k8s proxy to the cerebro service.
    """
    result = requests.Session()
    result.mount("http://", adapter)
    return result


class Cerebro:
    """API object for working with Cerebro."""

    def __init__(self, api):
        """Create a Cerebro instance.

        :param api: A Cerebro session handle for making HTTP requests with.
            See `cerebro_session` for a factory to create Cerebro session
            handles.
        """
        self._api = api

    def assignments(self, shard=None):
        """Return all assignments registered with Cerebro.

        :param shard: Filter for assignments to the one shard.
        """
        all_assignments = self._get_collection(Assignment)
        if not shard:
            yield from all_assignments
            return
        shard_id = shard.obj["id"]
        for assignment in all_assignments:
            if assignment.obj["shardID"] != shard_id:
                continue
            yield assignment

    def shards(self, serviceName=None, address=None):
        """Return all the shards registered with Cerebro.

        :param serviceName: Filter by the name of the service.
        :param address: Filter by the address of the shard.
        """
        # TODO: push the filter down to cerebro.
        all_shards = self._get_collection(Shard)
        if serviceName is None and address is None:
            yield from all_shards
            return
        for shard in all_shards:
            if serviceName and serviceName != shard.obj["serviceName"]:
                continue
            if address and address != shard.obj["address"]:
                continue
            yield shard

    def _get_collection(self, klass):
        r = self._api.get("http://cerebro-spring-service:8000" + klass.type_path)
        # TODO: retrieve expanded resources for efficiency
        try:
            r.raise_for_status()
        except Exception:
            # TODO: decorate the exception instead?
            logging.exception("Failed %s" % r.text)
            raise
        content = r.json()["content"]
        links = []
        for object in content:
            links.append(klass.type_path + "/" + object["id"])
        for link in links:
            instance = klass(self._api, {"link": link})
            instance.reload()
            yield instance


@contextlib.contextmanager
def patch_api(apiobject, new_api_session):
    """Use a different api session temporarily.

    For instance, to work around the k8s 1.5 issue where PATCH isn't proxied by
    the API server:

        with patch_api(shard, curl_session):
            shard.update()

    :param apiobject: An APIObject such as a Shard.
    :param new_api_session: A different cerebro_session() result to use while
        the context manager is active.
    :yield: None
    """
    old_session = apiobject._api
    try:
        apiobject._api = new_api_session
        yield None
    finally:
        apiobject._api = old_session


class APIObject:
    """An object in the Cerebro API.

    :attr obj: The underlying JSON object in dictionary form.
    :attr baseline: The last recieved copy of the JSON object from the server.
        Used to calculate semantic diffs to update collection fields like
        labels.
    """

    # The path to the collection of this object type.
    type_path = None  # type: str
    # What attributes should be merged from the server into the local
    # representation.
    merge_attrs = None  # type: Sequence[str]
    # What attributes need a ServiceStateMapUpdateRequest sent for PATCH ops
    map_attrs = ()  # type: Sequence[str]

    def __init__(self, api, obj_hash):
        """Create a Shard object.

        :param api: A Cerebro session handle for making HTTP requests with.
            See `cerebro_session` for a factory to create Cerebro session
            handles.
        :param obj_hash: The object a a Python dictionary.
        """
        self._api = api
        self.obj = obj_hash
        self.baseline = None

    def __str__(self):
        return "<%s %s>" % (
            type(self).__qualname__,
            json.dumps(self.obj, sort_keys=True),
        )

    def create(self):
        """Create the object in Cerebro."""
        result = self._update(self._api.post, self.type_path)
        self._refresh(result)
        logging.info("Created %s", self.obj["id"])
        return result

    def delete(self):
        """Delete this object from the service."""
        self_link = self.type_path + "/" + self.obj["id"]
        r = self._api.delete("http://cerebro-spring-service:8000" + self_link)
        try:
            r.raise_for_status()
        except Exception:
            # TODO: decorate the exception instead?
            logging.exception("Failed %s" % r.text)
            raise
        self.baseline = None
        self.obj = None
        logging.info("Deleted %s", self_link)

    def _update(self, method, endpoint, json=None):
        """Make a mutating call to the service.

        :param method: The method to make the call with, e.g. self._api.post.
        :param endpoint: The object to call, constructed from parent
        :param json: JSON to subject (in dict form). If not supplied uses
            self.obj.
        :return: The JSON result.
        """
        if json is None:
            json = self.obj

        r = method("http://cerebro-spring-service:8000" + endpoint, json=json)
        try:
            r.raise_for_status()
        except Exception:
            # TODO: decorate the exception instead?
            logging.exception("Failed %s" % r.text)
            raise
        return r.json()

    def _refresh(self, obj_hash):
        """Merge managed attributes of obj_hash into self."""
        try:
            for k in self.merge_attrs:
                self.obj[k] = obj_hash[k]
        except Exception:
            logging.exception("Could not merge %r into obj %r" % (obj_hash, self.obj))
            raise
        self.baseline = {}
        for k in self.merge_attrs:
            self.baseline[k] = deepcopy(obj_hash[k])

    def reload(self):
        """Reload the object from cerebro."""
        r = self._api.get("http://cerebro-spring-service:8000" + self.obj["link"])
        try:
            r.raise_for_status()
        except Exception:
            # TODO: decorate the exception instead?
            logging.exception("Failed %s" % r.text)
            raise
        result = r.json()
        self._refresh(result)

    def update(self):
        """Patch the service object with self.obj."""
        # Cerebro currently requires a separate ServiceStateMapUpdateRequest
        # request to update the labels hash.
        if self.map_attrs and self.baseline:
            adds = {}
            removals = {}
            for attr in self.map_attrs:
                baseline_attr = self.baseline.get(attr, {})
                new_attr = self.obj.get(attr, {})
                if baseline_attr == new_attr:
                    continue
                for k in set(baseline_attr) - set(new_attr):
                    removals.setdefault(attr, []).append(k)
                for k in set(new_attr) - set(baseline_attr):
                    adds.setdefault(attr, {})[k] = new_attr[k]
                for k in set(new_attr).intersection(set(baseline_attr)):
                    if new_attr[k] != baseline_attr[k]:
                        adds.setdefault(attr, {})[k] = new_attr[k]
            if adds or removals:
                # At least one map requires an update.
                patch = {
                    "kind": "com:vmware:xenon:common:ServiceStateMapUpdateRequest",  # noqa
                    "entriesToAdd": adds,
                    "keysToRemove": removals,
                }
                endpoint = self.type_path + "/" + self.obj["id"]
                self._update(self._api.patch, endpoint, json=patch)
        result = self._update(self._api.patch, endpoint)
        self._refresh(result)
        return result


class Shard(APIObject):
    """A shard in Cerebro.

    :attr obj: The shard object with keys:
    - serviceName: The Cerebro service name.
    - address: The address to send requests for this shard to.
    - labels: Labels for the service. Labels are k:v pairs that can be used
        by both users of cerebro and cerebro itself.
        Well known labels:
        "weight": "0" : disable the shard from scheduling
        "weight": other-numbers - weight of the shard for a biased coin toss on
            scheduling.
    - id: The unique id to the shard in the Cerebro API.
    """

    type_path = "/cerebro/resources/shards"
    merge_attrs = ("id", "address", "labels", "serviceName")
    map_attrs = ("labels",)

    def disable_assignments(self):
        """Disable assignments to the shard.

        Sets the weight to zero, performs an update, then waits for 30 seconds
        as a guard against concurrent assignments taking place.

        In practice, queries made after this function returns will see all
        newly made assignments.
        """
        if self.weight != "0":
            self.weight = "0"
            self.update()
            logging.info(
                "Shard %s Weight now %s" % (self.obj["id"], self.weight)
            )
            # wait a decent time for racing assignment requests to
            # complete. This is not strictly fenced, and may break
            # assignments but should do so extremely rarely.
            time.sleep(30)

    @property
    def weight(self):
        return self.obj["labels"]["weight"]

    @weight.setter
    def weight(self, value):
        self.obj["labels"]["weight"] = value


class Assignment(APIObject):
    """An assignment of a tenant to a shard in Cerebro.

    :attr obj: The assignment object with keys:
    - tenantId: The id of the tenant; typically a CSP org but service defined.
    - serviceName: The name of the service; must match the service name defined
      in the shard.
    - shardLink: The selflink of the shard object this assignment is for.
    """

    type_path = "/cerebro/resources/assignments"
    merge_attrs = ("id", "serviceName", "shardLink")


def get_csp_auth_token(session, csp_uri, user, secret):
    """Get a CSP client credentials token.

    The CSP endpoint can be found by hitting an auth login endpoint - e.g.
        https://api.testing.symphony-dev.com/authn/callback?
        redirect_url=https://api.testing.symphony-dev.com/discovery/
        and looking at the redirect path.

    :param session: A requests Session (or compatible) object.
    :param csp_uri: The CSP URI to request the token from
    :param user: The user to request the token for.
    :param secret: The secret to use to obtain the token.
    """
    return csp_lib.get_client_credentials_access_token(csp_uri, user, secret, session)


def upsert_shard(session, serviceName, address, weight="10", labels=None):
    """Helper for automation code which ensures a shard exists.

    If the shard already exists, no changes are made to it - though this may
    change depending on client feedback.

    :param session: A session from cerebro_session with which to communicate to
        cerebro.
    :param serviceName: The name of the service.
    :param address: The address that requests sharded onto this shard should be
        dispatched to.
    :param weight: The weight to assign to the shard.
    :param labels: Any additional / custom labels desired.
    :return: The created or existing shard.
    """
    # Workaround VSYM-14075, see also VSYM-14312
    # ---
    cerebro = Cerebro(session)

    def _deduped_shard():
        # We coerce to the lowest selfLink as this is consistent (assuming all
        # creation uses this codepath, there should also never be mass
        # assignments to migrate.
        shards = sorted(
            cerebro.shards(serviceName=serviceName, address=address),
            key=lambda s: s.obj["id"],
        )
        if not shards:
            return
        selected = shards[0]
        if len(shards) == 1:
            return selected
        serviceLink = selected.type_path + "/" + selected.obj["id"]
        logging.info(
            "Coercing all shards with address %s to shard %s", address, serviceLink
        )
        for shard in shards[1:]:
            # Step one: set shard to weight 0 to stop scheduling new
            # assignments while we remove it.
            shard.disable_assignments()
            # Step two: move all assignments for this shard to the target shard.
            assignments = list(cerebro.assignments(shard=shard))
            logging.info(
                "Updating %d assignments of %s ",
                len(assignments),
                shard.obj["id"],
            )
            for assignment in assignments:
                assignment.obj["shardID"] = serviceLink
                assignment.update()
                logging.info(
                    "assignment updated %s" % json.dumps(assignment.obj, sort_keys=True)
                )
            shard.delete()

    # ---
    existing = _deduped_shard()
    if existing:
        logging.info("Found existing shard %r", existing.obj["id"])
        return existing
    # None present thus far.
    if not labels:
        labels = {}
    else:
        labels = deepcopy(labels)
    labels["weight"] = weight
    for k, v in labels.items():
        if type(k) is not str or type(v) is not str:
            raise ValueError("Label %r:%r must be string->string." % (k, v))
    shard_obj = {"serviceName": serviceName, "address": address, "labels": labels}
    shard = Shard(session, shard_obj)
    shard.create()
    # Handle races on create - VSYM-14075
    return _deduped_shard()


def delete_shard(session, serviceName, address, delete_assignments=False):
    """Helper for automation code which deletes a shard.

    Lookup and delete a shard. If delete_assignments is true, assignments will
    be deleted, otherwise the presence of an assignment is an error and will
    raise a ValueError.

    During deletion, the weight is set to 0 and a 30 second wait is done to
    allow any racing assignments to take place.

    :param session: A session from cerebro_session with which to communicate to
        cerebro.
    :param serviceName: The name of the service.
    :param address: The address that requests sharded onto this shard should be
        dispatched to.
    """
    cerebro = Cerebro(session)
    existing = list(cerebro.shards(serviceName=serviceName, address=address))
    if not existing:
        return
    shard = existing[0]
    shard.disable_assignments()
    assignments = list(cerebro.assignments(shard=shard))
    if delete_assignments:
        for assignment in assignments:
            assignment.delete()
    else:
        if len(assignments):
            raise ValueError(
                "Shard %s still has %d assignments"
                % (shard.obj["id"], len(assignments))
            )
    shard.delete()


def assign_tenant_to_shard(session, tenantId, service_name, address):
    """assign a tanant to a shard

    This is typically used in service-injection test
    :param session: A session from cerebro_session with which to communicate to
        cerebro.
    :param tenantId: tenant ID
    :param service_name: service name
    :return: result of creation
    """
    cerebro = Cerebro(session)
    shards = list(cerebro.shards(serviceName=service_name, address=address))
    if not shards:
        raise ValueError("No shard was found with name %s", service_name)
    shard = shards[0]
    existing_assignment = list(cerebro.assignments(shard=shard))
    if existing_assignment:
        logging.info("Assignment already exist! %s", existing_assignment)
        return existing_assignment[0]
    assignment_obj = {
        "tenantId": tenantId,
        "serviceName": service_name,
        "shardId": shard.obj["shardId"],
    }
    assignment = Assignment(session, assignment_obj)
    return assignment.create()
