import logging
import requests
import yaml
from requests.auth import HTTPBasicAuth
from urllib.parse import urljoin
from scripts import polling
from scripts import cerebro_spring_lib

CSP_CLIENT_CREDENTIALS_AUTH_URI = "csp/gateway/am/api/auth/authorize"

CSP_ORG_OWNER_ROLE = "org_owner"
CSP_SERVICE_OWNER_ROLE = "service_owner"

CSP_AUTH_TOKEN_RESPONSE = "cspAuthToken"
REF_LINK_RESPONSE = "refLink"
REF_LINKS_RESPONSE = "refLinks"
SERVICE_LINK_RESPONSE = "serviceLink"

CSP_PLATFORM_OPERATOR_USERNAME = "csp@vmware.com"
CSP_PLATFORM_OPERATOR_PASSWORD = "VMware@123"

# CSP API Header Constants
CSP_AUTH_TOKEN_HEADER = "csp-auth-token"

# CSP URI Paths
CSP_AM_API_PATH = "/am/api"
CSP_GATEWAY_URI_PATH = "csp/gateway"
CSP_LOGIN_CONTAINER_URI = CSP_GATEWAY_URI_PATH + CSP_AM_API_PATH + "/login"
CSP_ORGS_PATH = "/orgs"
CSP_ORGS_URI = CSP_GATEWAY_URI_PATH + CSP_AM_API_PATH + CSP_ORGS_PATH
CSP_ROLES_PATH = "/roles"
CSP_USERS_URI = CSP_GATEWAY_URI_PATH + CSP_AM_API_PATH + "/users"
CSP_ORG_DEFINITION_URI = CSP_GATEWAY_URI_PATH + "/slc/api/definitions"
CSP_REGISTER_INSTANCE_URI = CSP_GATEWAY_URI_PATH + "/slc/api/lifecycle/register"
CSP_SERVICE_INVITE_URI = CSP_GATEWAY_URI_PATH + "/slc/api/service-invitations"
CSP_CLIENT_REGISTER_URI = CSP_GATEWAY_URI_PATH + CSP_AM_API_PATH + "/services/clients"
CSP_SERVICE_ROLES_PATH = "/service-roles"
CSP_REGISTRATION_URI = "/csp/registration"

# CSP Payload Body Constants
CSP_PAYLOAD_FIRST_NAME = "firstName"
CSP_PAYLOAD_LAST_NAME = "lastName"
CSP_PAYLOAD_USERNAME = "username"
CSP_PAYLOAD_PASSWORD = "password"
CSP_PAYLOAD_ROLE_NAMES_TO_ADD = "roleNamesToAdd"

# CSP Grant Types
CSP_GRANT_TYPE_AUTHORIZATION_CODE = "authorization_code"
CSP_GRANT_TYPE_REFRESH_TOKEN = "refresh_token"
CSP_GRANT_TYPE_CLIENT_CREDENTIALS = "client_credentials"


class Status409(Exception):
    """Permit handling 409 in future?"""


def _wait_for_200(method, *args, timeout_seconds=None, allowed_status=None, **kwargs):
    """Wait for a 200 success to a requests post."""
    if allowed_status is None:
        allowed_status = (200,)

    def _attempt():
        try:
            r = method(*args, **kwargs)
            if r.status_code not in allowed_status:
                logging.info("CSP result %s %d" % (args[0], r.status_code))
                if r.status_code == 409:
                    raise Status409(
                        "%s %s failed with %s:%r" % (args, kwargs, r.status_code, r.text)
                    )
                raise polling.Incomplete(
                    "%s %s failed with %s:%r" % (args, kwargs, r.status_code, r.text)
                )
            return r
        except Status409:
            raise
        except Exception as e:
            # TODO: make this narrower and make it specifically handle DNS
            # lookups (assuming such things raise from requests.post).
            raise polling.Incomplete(e)

    if timeout_seconds is not None:
        wait_kwargs = dict(timeout_seconds=timeout_seconds)
    else:
        wait_kwargs = {}
    return polling.wait_for(
        _attempt, "Requesting %s %s" % (args, kwargs), **wait_kwargs
    )


def _wait_for_json(*args, **kwargs):
    return _wait_for_200(*args, **kwargs).json()


def get_client_credentials_access_token(csp_uri, username, password, csp_session):
    """Returns the associated client_credentials access token.

    :param csp_uri: The CSP Host URI
    :param username: The client_credential's username
    :param password: The client_credential's password
    :param password: Optional 'requests' session to use
    """
    uri = urljoin(csp_uri, CSP_CLIENT_CREDENTIALS_AUTH_URI)
    logging.info(
        "Authenticating with CSP [uri=%s] as client_credentials [user=%s]",
        uri,
        username,
    )
    session = csp_session or requests.Session()
    body = "grant_type=client_credentials"

    json_result = _wait_for_json(
        session.post,
        uri,
        data=body,
        timeout_seconds=300,
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        auth=HTTPBasicAuth(username, password),
    )
    if "access_token" not in json_result:
        raise Exception(
            "'access_token' not returned in JSON response: %s" % json_result
        )
    return json_result["access_token"]


def login_as_csp_user(
        username, password, csp_session
):
    """Logs in as CSP user `username` at CSP Host `csp_uri`

    :param username: The username of the CSP user attempting to log in.
    :param password: The password of the CSP user attempting to log in.
    :param csp_session: A bound requests session - supercedes csp_uri, api and
        namespace. When supplied, any hostname given in an http method call should
        behave as a requests method call would, but routed to the CSP endpoint.
        See `new_csp_session()`.
    :return: The authorization token of the user.
    """
    login_as_csp_body = {
        CSP_PAYLOAD_USERNAME: username,
        CSP_PAYLOAD_PASSWORD: password,
    }
    json_result = _wait_for_json(
        csp_session.post,
        _api_url(CSP_LOGIN_CONTAINER_URI),
        json=login_as_csp_body,
        timeout_seconds=300,
    )

    if json_result[CSP_AUTH_TOKEN_RESPONSE] is not None:
        logging.info("Login as " + username + " successful.")

    return json_result[CSP_AUTH_TOKEN_RESPONSE]


def _api_url(suffix):
    return "http://csp-service/" + suffix


def new_csp_session(api, namespace):
    """Get a session for making CSP calls with.

    :param api: Pykube API instance.
    :param namespace: The namespace CSP is deployed in.
    """
    # TODO: if we keep this for any length of time, refactor the code from
    # cerebro_spring_lib into a common place.
    adapter = cerebro_spring_lib.KubernetesProxyAdapter(
        api, namespace, "csp-service", "csp-http", None
    )
    return cerebro_spring_lib.cerebro_session(adapter)


def create_csp_organization(
        csp_auth_token,
        organization_name,
        csp_session
):
    """Creates a CSP organization at CSP host `csp_uri` with organization name
    `organization_name`.

    :param csp_auth_token: The authorization token for the logged in user.
    :param organization_name: The organization name to search for.
    :param csp_session: A bound requests session - supercedes csp_uri, api and
        namespace. When supplied, any hostname given in an http method call should
        behave as a requests method call would, but routed to the CSP endpoint.
        See `new_csp_session()`.
    :return: The org_link of the created organization.
    """
    body = {"displayName": organization_name}
    json_result = _wait_for_json(
        csp_session.post,
        _api_url(CSP_ORGS_URI),
        headers={CSP_AUTH_TOKEN_HEADER: csp_auth_token},
        json=body,
        timeout_seconds=120,
    )
    if json_result[REF_LINK_RESPONSE] is not None:
        logging.info(
            "Organization %r successfully created at %s."
            % (organization_name, json_result[REF_LINK_RESPONSE])
        )

    return json_result[REF_LINK_RESPONSE]


def patch_csp_service_organization_with_service_owner_role(org_link, csp_session):
    """Patches the CSP service organization with `service_owner` role.

    :param org_link: The organization link of the organization being patched.
    :param csp_session: A bound requests session - supercedes csp_uri, api and
        namespace. When supplied, any hostname given in an http method call should
        behave as a requests method call would, but routed to the CSP endpoint.
        See `new_csp_session()`.
    """
    body = {CSP_PAYLOAD_ROLE_NAMES_TO_ADD: [CSP_SERVICE_OWNER_ROLE]}

    headers = {
        CSP_AUTH_TOKEN_HEADER: login_as_csp_user(
            CSP_PLATFORM_OPERATOR_USERNAME,
            CSP_PLATFORM_OPERATOR_PASSWORD,
            csp_session
        )
    }
    _wait_for_200(
        csp_session.patch,
        _api_url(org_link[1:] + CSP_ROLES_PATH),
        headers=headers,
        json=body,
        timeout_seconds=120,
        allowed_status=(200, 400),
    )


def create_csp_user(
        first_name,
        last_name,
        username,
        password,
        csp_session,
):
    """Creates a CSP user in the CSP host at `csp_uri`.

    :param first_name: The first name of the CSP user being created.
    :param last_name: The last name of the CSP user being created.
    :param username: The e-mail address of the CSP user being created.
    :param password: The password of the CSP user being created.
    :param csp_session: A bound requests session - supercedes csp_uri, api and
        namespace. When supplied, any hostname given in an http method call should
        behave as a requests method call would, but routed to the CSP endpoint.
        See `new_csp_session()`.
    """
    body = {
        CSP_PAYLOAD_FIRST_NAME: first_name,
        CSP_PAYLOAD_LAST_NAME: last_name,
        CSP_PAYLOAD_USERNAME: username,
        CSP_PAYLOAD_PASSWORD: password,
    }
    try:
        json_result = _wait_for_json(
            csp_session.post, _api_url(CSP_USERS_URI), json=body, timeout_seconds=120
        )
    except Status409:
        return
    if json_result[REF_LINK_RESPONSE] is not None:
        logging.info("User successfully created.")

    return json_result[REF_LINK_RESPONSE]


def patch_user_with_role(
        username, roles, org_link, csp_session
):
    """Patches the CSP user with role.

    :param username: The username of the user being patched.
    :param roles: The roles of user being patched
    :param org_link: The organization link of the organization being patched.
    :param csp_session: A bound requests session - supercedes csp_uri, api and
        namespace. When supplied, any hostname given in an http method call should
        behave as a requests method call would, but routed to the CSP endpoint.
        See `new_csp_session()`.
    """
    body = {CSP_PAYLOAD_ROLE_NAMES_TO_ADD: roles}
    org_id = org_link.split("/")[-1]

    headers = {
        CSP_AUTH_TOKEN_HEADER: login_as_csp_user(
            CSP_PLATFORM_OPERATOR_USERNAME,
            CSP_PLATFORM_OPERATOR_PASSWORD,
            csp_session=csp_session
        )
    }
    json_result = _wait_for_json(
        csp_session.patch,
        _api_url(
            CSP_USERS_URI
            + "/"
            + username
            + CSP_ORGS_PATH
            + "/"
            + org_id
            + CSP_ROLES_PATH
        ),
        headers=headers,
        json=body,
        timeout_seconds=120,
    )
    if json_result["roleNamesToAdd"] is not None:
        logging.info("Patch user with role successfully")
    return json_result["roleNamesToAdd"]


def create_service_definition(
        csp_auth_token, body, csp_session
):
    """Creates a service definition in the CSP host at `csp_uri`.

    :param csp_auth_token: The authorization token for the logged in user.
    :param body: The service definition YAML request body.
    :param csp_session: A bound requests session - supercedes csp_uri, api and
        namespace. When supplied, any hostname given in an http method call should
        behave as a requests method call would, but routed to the CSP endpoint.
        See `new_csp_session()`.
    :return: service definition link
    """
    json_result = _wait_for_json(
        csp_session.post,
        _api_url(CSP_ORG_DEFINITION_URI),
        headers={CSP_AUTH_TOKEN_HEADER: csp_auth_token, "content-type": "text/plain"},
        data=body,
        timeout_seconds=120,
    )
    if json_result[REF_LINK_RESPONSE] is not None:
        logging.info("Service definition successfully created.")
    return json_result[REF_LINK_RESPONSE]


def register_instance(
        instance_name,
        service_definition_link,
        public_url,
        csp_auth_token,
        csp_session,
):
    """Register instance in the CSP host at `csp_uri`.

    :param instance_name: The service instance_name will be registered
    :param service_definition_link: The instance service_definition_link will be reigstered
    :param public_url: The instance public_url will be reigstered
    :param csp_auth_token: The authorization token for the logged in user.
    :param csp_session: A bound requests session - supercedes csp_uri, api and
        namespace. When supplied, any hostname given in an http method call should
        behave as a requests method call would, but routed to the CSP endpoint.
        See `new_csp_session()`.
    :return: registration service link
    """
    body = {
        "instanceName": instance_name,
        "serviceDefinitionLink": service_definition_link,
        "publicUrl": public_url,
    }
    json_result = _wait_for_json(
        csp_session.post,
        _api_url(CSP_REGISTER_INSTANCE_URI),
        headers={CSP_AUTH_TOKEN_HEADER: csp_auth_token},
        json=body,
        timeout_seconds=120,
    )
    if json_result[SERVICE_LINK_RESPONSE] is not None:
        logging.info("Register instance successfully.")

    return json_result[SERVICE_LINK_RESPONSE]


def create_service_invitations(
        service_definition_link,
        csp_auth_token,
        csp_session
):
    """Create service invitations in the CSP host at `csp_uri`.

    :param service_definition_link: The instance service_definition_link will be reigstered
    :param csp_auth_token: The authorization token for the logged in user.
    :param csp_session: A bound requests session - supercedes csp_uri, api and
        namespace. When supplied, any hostname given in an http method call should
        behave as a requests method call would, but routed to the CSP endpoint.
        See `csp_session()`.
    :return: service invitation link
    """
    body = {
        "serviceDefinitionLink": service_definition_link,
        "numberOfInvitationsToGenerate": 1,
        "context": {},
        "isFundMandatory": False,
    }
    json_result = _wait_for_json(
        csp_session.post,
        _api_url(CSP_SERVICE_INVITE_URI),
        headers={CSP_AUTH_TOKEN_HEADER: csp_auth_token},
        json=body,
        timeout_seconds=120,
    )
    if json_result[REF_LINKS_RESPONSE] is not None:
        return (
            json_result[REF_LINKS_RESPONSE][0].split("=")[1].replace("/csp", "")
        )


def register_csp_client(
        client_id,
        client_secret,
        csp_auth_token,
        service_definition_link,
        grant_types,
        csp_session,
        redirect_uri=""
):
    """Register csp client to the CSP host at `csp_uri`.

    :param client_id: The client id
    :param client_secret: The client secret
    :param csp_auth_token: The authorization token for the logged in user.
    :param service_definition_link: The service invitation link
    :param grant_types: The client types will be granted
    :param csp_session: A bound requests session - supercedes csp_uri, api and
        namespace. When supplied, any hostname given in an http method call should
        behave as a requests method call would, but routed to the CSP endpoint.
        See `csp_session()`.
    """
    body = {
        "id": client_id,
        "secret": client_secret,
        "grantTypes": grant_types,
        "resourceLink": service_definition_link,
        "redirectUri": redirect_uri,
        "accessTokenTTL": 1800,
        "refreshTokenTTL": 18000,
    }
    json_result = _wait_for_json(
        csp_session.post,
        _api_url(CSP_CLIENT_REGISTER_URI),
        headers={CSP_AUTH_TOKEN_HEADER: csp_auth_token},
        json=body,
        timeout_seconds=120,
    )
    if json_result[REF_LINK_RESPONSE] is not None:
        logging.info("Register csp client successfully")
    return json_result[REF_LINK_RESPONSE]


def add_members_to_org(
        usernames,
        org_roles,
        org_definition_link,
        csp_auth_token,
        csp_session
):
    """Add members to org in the CSP host at `csp_uri`.

    :param usernames: The instance service_definition_link will be reigstered
    :param org_roles: The roles granted for the user
    :param org_definition_link: The org definition link
    :param csp_auth_token: The authorization token for the logged in user.
    :param csp_session: A bound requests session - supercedes csp_uri, api and
        namespace. When supplied, any hostname given in an http method call should
        behave as a requests method call would, but routed to the CSP endpoint.
        See `csp_session()`.
    """
    body = {"usernames": usernames, "orgRoleName": org_roles}
    json_result = _wait_for_json(
        csp_session.post,
        _api_url(org_definition_link[1:] + "/invitations"),
        headers={CSP_AUTH_TOKEN_HEADER: csp_auth_token},
        json=body,
        timeout_seconds=120,
        allowed_status=(200, 202),
    )
    if json_result["usernames"] is not None:
        logging.info("add members to org successfully")
    return json_result["usernames"]


def redeem_invitation(
        org_definition_link,
        service_invitation_link,
        csp_auth_token,
        csp_session
):
    """Redeem invitation in the CSP host at `csp_uri`.

    :param org_definition_link: The org definition link
    :param service_invitation_link: The service invitation link
    :param csp_auth_token: The authorization token for the logged in user.
    :param csp_session: A bound requests session - supercedes csp_uri, api and
        namespace. When supplied, any hostname given in an http method call should
        behave as a requests method call would, but routed to the CSP endpoint.
        See `csp_session()`.
    """
    body = {"orgLink": org_definition_link, "invitationAction": "REDEEM"}
    json_result = _wait_for_json(
        csp_session.patch,
        _api_url(CSP_GATEWAY_URI_PATH + service_invitation_link),
        headers={CSP_AUTH_TOKEN_HEADER: csp_auth_token},
        json=body,
        timeout_seconds=120,
    )
    if json_result[REF_LINK_RESPONSE] is not None:
        logging.info("Redeem user successfully")
    return json_result[REF_LINK_RESPONSE]


def patch_user_with_service_role(
        username,
        roles,
        service_definition_link,
        org_link,
        csp_session
):
    """Patches the CSP user with service role.

    :param username: The username of the user being patched.
    :param roles: The roles of user being patched
    :param service_definition_link: The service definition link
    :param org_link: The organization link of the organization being patched.
    :param csp_session: A bound requests session - supercedes csp_uri, api and
        namespace. When supplied, any hostname given in an http method call should
        behave as a requests method call would, but routed to the CSP endpoint.
        See `csp_session()`.
    """
    body = {
        "serviceDefinitionLink": service_definition_link,
        CSP_PAYLOAD_ROLE_NAMES_TO_ADD: roles,
    }
    org_id = org_link.split("/")[-1]
    headers = {
        CSP_AUTH_TOKEN_HEADER: login_as_csp_user(
            CSP_PLATFORM_OPERATOR_USERNAME,
            CSP_PLATFORM_OPERATOR_PASSWORD,
            csp_session,
        )
    }
    json_result = _wait_for_json(
        csp_session.patch,
        _api_url(
            CSP_USERS_URI
            + "/"
            + username
            + CSP_ORGS_PATH
            + "/"
            + org_id
            + CSP_SERVICE_ROLES_PATH
        ),
        headers=headers,
        json=body,
        timeout_seconds=120,
    )

    if json_result["roleNamesToAdd"] is not None:
        logging.info("Patch user with service role successfully")


# Dirty port from automation service
def bootstrap_registration(csp_session, data):
    token = login_as_csp_user(
        data["cspUser"],
        data["cspPassword"],
        csp_session
    )

    org_ref = create_csp_organization(
        token,
        data["serviceDisplayName"],
        csp_session
    )

    create_csp_user(
        data["serviceOwnerUser"]["firstName"],
        data["serviceOwnerUser"]["lastName"],
        data["serviceOwnerUser"]["username"],
        data["serviceOwnerUser"]["password"],
        csp_session
        # email = data["serviceOwnerUser"]["email // Implement if required
    )

    service_definition_data = {
        "name": data["serviceName"],
        "display-name": data["serviceDisplayName"],
        "desc-short": data["serviceShortDesc"],
        "desc-long": data["serviceLongDesc"],
        "internal": False,
        "health-check-path": data["healthCheckPath"],
        "service-urls": {
            "service-home": data["serviceUrls"]["service-home"],
            "terms-of-service": data["serviceUrls"]["terms-of-service"]
        },
        "gated": True,
        "org-id": org_ref,
        "service-icon": data["serviceIcon"],
        "service-nav-bar-icon": data["serviceNavBarIcon"],
        "service-roles": data["serviceUserRoles"]
    }

    service_definition_ref = create_service_definition(
        token,
        yaml.dump(service_definition_data),
        csp_session
    )

    register_instance(
        data["serviceName"],
        service_definition_ref,
        data["serviceUrls"]["service-home"],
        token,
        csp_session,
    )

    patch_user_with_role(
        data["serviceOwnerUser"]["username"],
        [CSP_ORG_OWNER_ROLE],
        org_ref,
        csp_session
    )

    for client_data in data["authClientConfigs"]:
        register_csp_client(
            client_data["id"],
            client_data["secret"],
            token,
            service_definition_ref,
            [
                CSP_GRANT_TYPE_AUTHORIZATION_CODE,
                CSP_GRANT_TYPE_REFRESH_TOKEN,
                CSP_GRANT_TYPE_CLIENT_CREDENTIALS
            ],
            csp_session,
            client_data["redirectUri"]
        )

    service_invitation_link = create_service_invitations(
        service_definition_ref,
        token,
        csp_session
    )

    redeem_invitation(
        org_ref,
        service_invitation_link,
        token,
        csp_session
    )

    patch_user_with_service_role(
        data["serviceOwnerUser"]["username"],
        [data["serviceUserRoles"][0]["name"]],  # Yes, this is what's written in automation service
        service_definition_ref,
        org_ref,
        csp_session
    )
