import logging
import time


class Incomplete(Exception):
    """Used to indicate incomplete predicates."""


def wait_for(check, msg, timeout_seconds=600, period_seconds=10):
    """Wait for check to be successful.

    :param check: A callable no-args function. Should raise
        Incomplete if the desired predicate has not been met.
    :param msg: A constant to log while waiting.
    :param timeout_seconds: The maximum time to wait.
    :param period_seconds: How frequently to poll.
    :return: The return value of check.
    :raises Incomplete: If the timeout is expired, the
        last exception raised from check is propogated.
    """
    max_time = time.time() + timeout_seconds
    now = time.time()
    last_err = None
    count = 0
    # While rather than range, to deal with slow APIs.
    while now < max_time:
        count += 1
        logging.info("Waiting for %s: attempt [%s]", msg, count)
        try:
            return check()
        except Incomplete as e:
            last_err = e
        time.sleep(period_seconds)
        now = time.time()
    logging.error("Timeout waiting for %s", msg)
    raise last_err
